import logging
from openstack.connection import Connection
from osi_dump import util
from osi_dump.importer.role_assignment.role_assignment_importer import RoleAssignmentImporter
from osi_dump.importer.role_assignment.openstack_role_assignment_importer import OpenStackRoleAssignmentImporter
from osi_dump.exporter.role_assignment.role_assignment_exporter import RoleAssignmentExporter
from osi_dump.exporter.role_assignment.excel_role_assignment_exporter import ExcelRoleAssignmentExporter

logger = logging.getLogger(__name__)

class RoleAssignmentBatchHandler:
    def __init__(self):
        self._importer_exporter_list: list[tuple[RoleAssignmentImporter, RoleAssignmentExporter]] = []

    def add_importer_exporter_from_openstack_connections(
        self, connections: list[Connection], output_file: str
    ):
        for connection in connections:
            importer = OpenStackRoleAssignmentImporter(connection)
            sheet_name_prefix = f"{util.extract_hostname(connection.auth['auth_url'])}-role"
            exporter = ExcelRoleAssignmentExporter(
                sheet_name_prefix=sheet_name_prefix, output_file=output_file
            )
            self.add_importer_exporter(importer=importer, exporter=exporter)

    def add_importer_exporter(
        self, importer: RoleAssignmentImporter, exporter: RoleAssignmentExporter
    ):
        self._importer_exporter_list.append((importer, exporter))

    def process(self):
        for importer, exporter in self._importer_exporter_list:
            try:
                exporter.export_role_assignments(importer=importer)
            except Exception as e:
                logger.warning(e)
                logger.warning("Skipping...")


import logging
import pandas as pd
from osi_dump import util
from .role_assignment_exporter import RoleAssignmentExporter
from osi_dump.importer.role_assignment.role_assignment_importer import RoleAssignmentImporter

logger = logging.getLogger(__name__)

class ExcelRoleAssignmentExporter(RoleAssignmentExporter):
    def __init__(self, sheet_name_prefix: str, output_file: str):
        self.sheet_name_prefix = sheet_name_prefix
        self.output_file = output_file

    def export_role_assignments(self, importer: RoleAssignmentImporter):
        # Sheet 1: Raw User Role Assignments
        df_user = pd.json_normalize(
            (role.model_dump() for role in importer.get_user_roles())
        )
        if not df_user.empty:
            self._export_to_sheet(df_user, f"{self.sheet_name_prefix}-user")
            
        # Sheet 2: Raw Group Role Assignments
        df_group = pd.json_normalize(
            (role.model_dump() for role in importer.get_group_roles())
        )
        if not df_group.empty:
            self._export_to_sheet(df_group, f"{self.sheet_name_prefix}-group")

    def _export_to_sheet(self, df: pd.DataFrame, sheet_name: str):
        logger.info(f"Exporting data to sheet: {sheet_name}")
        try:
            util.export_data_excel(self.output_file, sheet_name=sheet_name, df=df)
            logger.info(f"Exported data to sheet: {sheet_name}")
        except Exception as e:
            logger.warning(f"Exporting to {sheet_name} error: {e}")


from abc import ABC, abstractmethod
from typing import Generator
from osi_dump.model.role_assignment import UserRoleAssignment, GroupRoleAssignment

class RoleAssignmentImporter(ABC):
    @abstractmethod
    def get_user_roles(self) -> Generator[UserRoleAssignment, None, None]:
        pass

    @abstractmethod
    def get_group_roles(self) -> Generator[GroupRoleAssignment, None, None]:
        pass
    
#    @abstractmethod
#    def calculate_effective_roles(self) -> Generator[EffectiveUserRole, None, None]:
#        pass


from typing import Optional, List
from pydantic import BaseModel, ConfigDict

class SecurityGroupRule(BaseModel):
    """Represents a single rule within a security group."""
    rule_id: str
    direction: str
    protocol: Optional[str]
    ethertype: str
    port_range: Optional[str]
    remote_ip_prefix: Optional[str]
    remote_group_id: Optional[str]

class SecurityGroup(BaseModel):
    """Represents an OpenStack Security Group."""
    model_config = ConfigDict(strict=True)

    security_group_id: str
    name: str
    project_id: Optional[str]
    description: Optional[str]
    rules: List[SecurityGroupRule]


from abc import ABC, abstractmethod
from typing import Generator
from osi_dump.model.security_group import SecurityGroup

class SecurityGroupImporter(ABC):
    @abstractmethod
    def import_security_groups(self) -> Generator[SecurityGroup, None, None]: 
        pass


from abc import ABC, abstractmethod
from typing import Generator
from osi_dump.model.instance import Instance

class InstanceExporter(ABC):
    @abstractmethod
    def export_instances(self, instances: Generator[Instance, None, None]):
        pass


import logging
from openstack.connection import Connection
from osi_dump import util
from osi_dump.importer.security_group.security_group_importer import SecurityGroupImporter
from osi_dump.exporter.security_group.security_group_exporter import SecurityGroupExporter
from osi_dump.importer.security_group.openstack_security_group_importer import OpenStackSecurityGroupImporter
from osi_dump.exporter.security_group.excel_security_group_exporter import ExcelSecurityGroupExporter

logger = logging.getLogger(__name__)

class SecurityGroupBatchHandler:
    def __init__(self):
        self._importer_exporter_list: list[tuple[SecurityGroupImporter, SecurityGroupExporter]] = []

    def add_importer_exporter_from_openstack_connections(
        self, connections: list[Connection], output_file: str
    ):
        for connection in connections:
            importer = OpenStackSecurityGroupImporter(connection)
            sheet_name = f"{util.extract_hostname(connection.auth['auth_url'])}-sec-group"
            exporter = ExcelSecurityGroupExporter(
                sheet_name=sheet_name, output_file=output_file
            )
            self._importer_exporter_list.append((importer, exporter))

    def process(self):
        for importer, exporter in self._importer_exporter_list:
            try:
                # importer get generators
                security_groups_generator = importer.import_security_groups()
                
                # exporter uses generators
                exporter.export_security_groups(security_groups=security_groups_generator)

            except Exception as e:
                logger.warning(e)
                logger.warning("Skipping security group export...")


import logging
from typing import Generator
from openstack.connection import Connection
from openstack.network.v2.security_group import SecurityGroup as OSSecurityGroup

from osi_dump.importer.security_group.security_group_importer import SecurityGroupImporter
from osi_dump.model.security_group import SecurityGroup, SecurityGroupRule

logger = logging.getLogger(__name__)

class OpenStackSecurityGroupImporter(SecurityGroupImporter):
    def __init__(self, connection: Connection):
        self.connection = connection

    def import_security_groups(self) -> Generator[SecurityGroup, None, None]:
        logger.info(f"Importing security groups for {self.connection.auth['auth_url']}")
        try:
            os_sec_groups_iterator = self.connection.network.security_groups()

            for os_sec_group in os_sec_groups_iterator:
                yield self._get_sec_group_info(os_sec_group)

        except Exception as e:
            logger.error(f"Cannot fetch security groups for {self.connection.auth['auth_url']}: {e}")
            return 

        logger.info(f"Finished importing security groups for {self.connection.auth['auth_url']}")


    def _get_sec_group_info(self, os_sec_group: OSSecurityGroup) -> SecurityGroup:
        rules = []
        if hasattr(os_sec_group, 'security_group_rules'):
            for rule in os_sec_group.security_group_rules:
                port_range = None
                if rule.get('port_range_min') is not None and rule.get('port_range_max') is not None:
                    port_range = f"{rule['port_range_min']}-{rule['port_range_max']}"

                rules.append(SecurityGroupRule(
                    rule_id=rule['id'],
                    direction=rule['direction'],
                    protocol=rule.get('protocol'),
                    ethertype=rule['ethertype'],
                    port_range=port_range,
                    remote_ip_prefix=rule.get('remote_ip_prefix'),
                    remote_group_id=rule.get('remote_group_id')
                ))
        
        return SecurityGroup(
            security_group_id=os_sec_group.id,
            name=os_sec_group.name,
            project_id=os_sec_group.project_id,
            description=os_sec_group.description,
            rules=rules
        )


import pandas as pd
import logging
from typing import Generator

from osi_dump import util
from osi_dump.exporter.instance.instance_exporter import InstanceExporter
from osi_dump.model.instance import Instance

logger = logging.getLogger(__name__)

class ExcelInstanceExporter(InstanceExporter):
    def __init__(self, sheet_name: str, output_file: str):
        self.sheet_name = sheet_name
        self.output_file = output_file

    def export_instances(self, instances: Generator[Instance, None, None]):
        df = pd.DataFrame(instance.model_dump() for instance in instances)

        if df.empty:
            logger.info(f"No instances to export for {self.sheet_name}")
            return
        df.sort_values(by='instance_name', inplace=True, na_position='last')

        logger.info(f"Exporting instances for {self.sheet_name}")
        try:
            util.export_data_excel(self.output_file, sheet_name=self.sheet_name, df=df)
            logger.info(f"Exported instances for {self.sheet_name}")
        except Exception as e:
            logger.warning(f"Exporting instances for {self.sheet_name} error: {e}")


import logging
from typing import Generator
from openstack.connection import Connection
from openstack.compute.v2.server import Server
from openstack.compute.v2.flavor import Flavor as OSFlavor

from osi_dump.importer.instance.instance_importer import InstanceImporter
from osi_dump.model.instance import Instance

logger = logging.getLogger(__name__)


class OpenStackInstanceImporter(InstanceImporter):
    def __init__(self, connection: Connection):
        self.connection = connection

    def import_instances(self) -> Generator[Instance, None, None]:
        """Import instances information from Openstack as a generator."""

        logger.info(f"Importing instances for {self.connection.auth['auth_url']}")
        try:
            server_iterator = self.connection.compute.servers(details=True, all_projects=True)
            
            for server in server_iterator:
                yield self._get_instance_info(server)

        except Exception as e:
            logger.error(f"Cannot fetch instances for {self.connection.auth['auth_url']}: {e}")
            return 
        
        logger.info(f"Finished importing instances for {self.connection.auth['auth_url']}")

    def _get_instance_info(self, server: Server) -> Instance:
        project_name = None
        project_id = None
        project = None 
        try:
            project = self.connection.identity.get_project(server.project_id)
            project_name = project.name
            project_id = project.id
        except Exception as e:
            logger.warning(
                f"Unable to obtain project name for instance: {server.name}: {e}"
            )

        domain_name = None
        if project: 
            try:
                domain = self.connection.identity.get_domain(project.domain_id)
                domain_name = domain.name
            except Exception as e:
                logger.warning(
                    f"Unable to obtain domain name for instance {server.name}: {e}"
                )

        private_v4_ips = []
        floating_ip = None

        try:
            for ips in server.addresses.values():
                for ip in ips:
                    if ip["OS-EXT-IPS:type"] == "fixed":
                        private_v4_ips.append(ip["addr"])
                    elif ip["OS-EXT-IPS:type"] == "floating":
                        floating_ip = ip["addr"]
        except Exception as e:
            logger.warning(
                f"Unable to obtain IP address information for instance {server.name}: {e}"
            )

        vgpus = None
        vgpu_type = None
        vgpu_metadata_property = "pci_passthrough:alias"

        try:
            flavor: OSFlavor = self.connection.get_flavor(
                name_or_id=server.flavor["id"]
            )
            if flavor and flavor.extra_specs:
                vgpu_prop: str = flavor.extra_specs.get(vgpu_metadata_property)
                if vgpu_prop:
                    vgpu_props = vgpu_prop.split(":")
                    vgpu_type = vgpu_props[0]
                    vgpus = int(vgpu_props[1])
        except Exception:
            pass
        
        image_id = server.image.get("id") if server.image else None
        flavor_id = server.flavor.get("id") if server.flavor else None
        
        return Instance(
            instance_id=server.id,
            instance_name=server.name,
            project_id=project_id,
            project_name=project_name,
            domain_name=domain_name,
            private_v4_ips=private_v4_ips,
            floating_ip=floating_ip,
            status=server.status,
            hypervisor=server.hypervisor_hostname,
            ram=server.flavor["ram"],
            vcpus=server.flavor["vcpus"],
            created_at=server.created_at,
            updated_at=server.updated_at,
            user_id=server.user_id,
            vgpus=vgpus,
            vgpu_type=vgpu_type,
            image_id=image_id, 
            flavor_id=flavor_id
        )


import logging
from openstack.connection import Connection
from openstack.identity.v3.role_assignment import RoleAssignment as OSRoleAssignment
from typing import Generator

from .role_assignment_importer import RoleAssignmentImporter
from osi_dump.model.role_assignment import UserRoleAssignment, GroupRoleAssignment
from osi_dump.api.keystone import get_users

logger = logging.getLogger(__name__)

class OpenStackRoleAssignmentImporter(RoleAssignmentImporter):
    def __init__(self, connection: Connection):
        self.connection = connection
        self.users = {}
        self.roles = {}
        self.groups = {}
        self._raw_assignments = None 

    def _fetch_metadata(self):
        """cache metadata"""
        if self._raw_assignments is not None:
            return 

        logger.info(f"Pre-caching metadata for role assignments on {self.connection.auth['auth_url']}")
        
        try:
            os_users = get_users(self.connection)
            for os_user in os_users: self.users[os_user["id"]] = os_user
        except Exception as e:
            logger.error(f"Could not fetch users: {e}")

        try:
            os_roles = self.connection.identity.roles()
            for os_role in os_roles: self.roles[os_role.id] = os_role.name
        except Exception as e:
            logger.error(f"Could not fetch roles: {e}")
            
        try:
            os_groups = list(self.connection.identity.groups())
            for group in os_groups: self.groups[group.id] = group.name
        except Exception as e:
            logger.error(f"Could not fetch groups: {e}")

        try:
            self._raw_assignments = list(self.connection.identity.role_assignments())
        except Exception as e:
            logger.error(f"Could not fetch role assignments: {e}")
            self._raw_assignments = []

    def get_user_roles(self) -> Generator[UserRoleAssignment, None, None]:
        self._fetch_metadata()
        for assignment in self._raw_assignments:
            if assignment.user:
                user_id = assignment.user.get('id')
                user_info = self.users.get(user_id, {})
                role_id = assignment.role.get('id') if assignment.role else None
                yield UserRoleAssignment(
                    user_id=user_id,
                    user_name=user_info.get('name'),
                    role_id=role_id,
                    role_name=self.roles.get(role_id),
                    scope=assignment.scope,
                    enabled=user_info.get('enabled'),
                    password_expires_at=user_info.get('password_expires_at'),
                    options=user_info.get('options')
                )

    def get_group_roles(self) -> Generator[GroupRoleAssignment, None, None]:
        self._fetch_metadata()
        for assignment in self._raw_assignments:
            if assignment.group:
                group_id = assignment.group.get('id')
                role_id = assignment.role.get('id') if assignment.role else None
                yield GroupRoleAssignment(
                    group_id=group_id,
                    group_name=self.groups.get(group_id),
                    role_id=role_id,
                    role_name=self.roles.get(role_id),
                    scope=assignment.scope
                )


import logging
import pandas as pd
from typing import Generator
from osi_dump import util
from osi_dump.exporter.security_group.security_group_exporter import SecurityGroupExporter
from osi_dump.model.security_group import SecurityGroup

logger = logging.getLogger(__name__)

class ExcelSecurityGroupExporter(SecurityGroupExporter):
    def __init__(self, sheet_name: str, output_file: str):
        self.sheet_name = sheet_name
        self.output_file = output_file

    def export_security_groups(self, security_groups: Generator[SecurityGroup, None, None]):
        
        data_generator = (sg.model_dump() for sg in security_groups)

        df = pd.json_normalize(
            data_generator,
            record_path='rules',
            meta=['security_group_id', 'name', 'project_id', 'description'],
            record_prefix='rule.'
        )

        if df.empty:
            logger.info(f"No security groups to export for {self.sheet_name}")
            return
            
        logger.info(f"Exporting security groups for {self.sheet_name}")
        try:
            util.export_data_excel(self.output_file, sheet_name=self.sheet_name, df=df)
            logger.info(f"Exported security groups for {self.sheet_name}")
        except Exception as e:
            logger.warning(f"Exporting security groups for {self.sheet_name} error: {e}")


from abc import ABC, abstractmethod
from typing import Generator
from osi_dump.model.instance import Instance

class InstanceImporter(ABC):
    @abstractmethod
    def import_instances(self) -> Generator[Instance, None, None]:
        pass


from abc import ABC, abstractmethod
from osi_dump.importer.role_assignment.role_assignment_importer import RoleAssignmentImporter

class RoleAssignmentExporter(ABC):
    @abstractmethod
    def export_role_assignments(self, importer: RoleAssignmentImporter):
        pass


from abc import ABC, abstractmethod
from typing import Generator
from osi_dump.model.security_group import SecurityGroup

class SecurityGroupExporter(ABC):
    @abstractmethod
    def export_security_groups(self, security_groups: Generator[SecurityGroup, None, None]):
        pass


from typing import Optional
from pydantic import BaseModel, ConfigDict

class UserRoleAssignment(BaseModel):
    model_config = ConfigDict(strict=True)
    user_id: str
    user_name: Optional[str]
    role_id: str
    role_name: Optional[str]
    scope: dict
    enabled: Optional[bool]
    password_expires_at: Optional[str]
    options: Optional[dict]

class GroupRoleAssignment(BaseModel):
    model_config = ConfigDict(strict=True)
    group_id: str
    group_name: Optional[str]
    role_id: str
    role_name: Optional[str]
    scope: dict

#class EffectiveUserRole(BaseModel):
#    model_config = ConfigDict(strict=True)
#    user_id: str
#    user_name: Optional[str]
#    role_id: str
#    role_name: Optional[str]
#    scope: dict
#    inherited_from_group: Optional[str]